#!/usr/bin/python3
# Plot GNU C Library string microbenchmark output.
# Copyright (C) 2019-2021 Free Software Foundation, Inc.
# This file is part of the GNU C Library.
#
# The GNU C Library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# The GNU C Library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the GNU C Library; if not, see
# <https://www.gnu.org/licenses/>.
"""Plot string microbenchmark results.

Given a benchmark results file in JSON format and a benchmark schema file,
plot the benchmark timings in one of the available representations.

Separate figure is generated and saved to a file for each 'results' array
found in the benchmark results file. Output filenames and plot titles
are derived from the metadata found in the benchmark results file.
"""
import argparse
from collections import defaultdict
import json
import matplotlib as mpl
import numpy as np
import os

try:
    import jsonschema as validator
except ImportError:
    print("Could not find jsonschema module.")
    raise

# Use pre-selected markers for plotting lines to improve readability
markers = [".", "x", "^", "+", "*", "v", "1", ">", "s"]

# Benchmark variants for which the x-axis scale should be logarithmic
log_variants = {"powers of 2"}


def gmean(numbers):
    """Compute geometric mean.

    Args:
        numbers: 2-D list of numbers
    Return:
        numpy array with geometric means of numbers along each column
    """
    a = np.array(numbers, dtype=np.complex)
    means = a.prod(0) ** (1.0 / len(a))
    return np.real(means)


def relativeDifference(x, x_reference):
    """Compute per-element relative difference between each row of
       a matrix and an array of reference values.

    Args:
        x: numpy matrix of shape (n, m)
        x_reference: numpy array of size m
    Return:
        relative difference between rows of x and x_reference (in %)
    """
    abs_diff = np.subtract(x, x_reference)
    return np.divide(np.multiply(abs_diff, 100.0), x_reference)


def plotTime(timings, routine, bench_variant, title, outpath):
    """Plot absolute timing values.

    Args:
        timings: timings to plot
        routine: benchmarked string routine name
        bench_variant: top-level benchmark variant name
        title: figure title (generated so far)
        outpath: output file path (generated so far)
    Return:
        y: y-axis values to plot
        title_final: final figure title
        outpath_final: file output file path
    """
    y = timings
    plt.figure()

    if not args.values:
        plt.axes().yaxis.set_major_formatter(plt.NullFormatter())

    plt.ylabel("timing")
    title_final = "%s %s benchmark timings\n%s" % \
                  (routine, bench_variant, title)
    outpath_final = os.path.join(args.outdir, "%s_%s_%s%s" % \
                    (routine, args.plot, bench_variant, outpath))

    return y, title_final, outpath_final


def plotRelative(timings, all_timings, routine, ifuncs, bench_variant,
                 title, outpath):
    """Plot timing values relative to a chosen ifunc

    Args:
        timings: timings to plot
        all_timings: all collected timings
        routine: benchmarked string routine name
        ifuncs: names of ifuncs tested
        bench_variant: top-level benchmark variant name
        title: figure title (generated so far)
        outpath: output file path (generated so far)
    Return:
        y: y-axis values to plot
        title_final: final figure title
        outpath_final: file output file path
    """
    # Choose the baseline ifunc
    if args.baseline:
        baseline = args.baseline.replace("__", "")
    else:
        baseline = ifuncs[0]

    baseline_index = ifuncs.index(baseline)

    # Compare timings against the baseline
    y = relativeDifference(timings, all_timings[baseline_index])

    plt.figure()
    plt.axhspan(-args.threshold, args.threshold, color="lightgray", alpha=0.3)
    plt.axhline(0, color="k", linestyle="--", linewidth=0.4)
    plt.ylabel("relative timing (in %)")
    title_final = "Timing comparison against %s\nfor %s benchmark, %s" % \
                  (baseline, bench_variant, title)
    outpath_final = os.path.join(args.outdir, "%s_%s_%s%s" % \
                    (baseline, args.plot, bench_variant, outpath))

    return y, title_final, outpath_final


def plotMax(timings, routine, bench_variant, title, outpath):
    """Plot results as percentage of the maximum ifunc performance.

    The optimal ifunc is computed on a per-parameter-value basis.
    Performance is computed as 1/timing.

    Args:
        timings: timings to plot
        routine: benchmarked string routine name
        bench_variant: top-level benchmark variant name
        title: figure title (generated so far)
        outpath: output file path (generated so far)
    Return:
        y: y-axis values to plot
        title_final: final figure title
        outpath_final: file output file path
    """
    perf = np.reciprocal(timings)
    max_perf = np.max(perf, axis=0)
    y = np.add(100.0, relativeDifference(perf, max_perf))

    plt.figure()
    plt.axhline(100.0, color="k", linestyle="--", linewidth=0.4)
    plt.ylabel("1/timing relative to max (in %)")
    title_final = "Performance comparison against max for %s\n%s " \
                  "benchmark, %s" % (routine, bench_variant, title)
    outpath_final = os.path.join(args.outdir, "%s_%s_%s%s" % \
                    (routine, args.plot, bench_variant, outpath))

    return y, title_final, outpath_final


def plotThroughput(timings, params, routine, bench_variant, title, outpath):
    """Plot throughput.

    Throughput is computed as the varied parameter value over timing.

    Args:
        timings: timings to plot
        params: varied parameter values
        routine: benchmarked string routine name
        bench_variant: top-level benchmark variant name
        title: figure title (generated so far)
        outpath: output file path (generated so far)
    Return:
        y: y-axis values to plot
        title_final: final figure title
        outpath_final: file output file path
    """
    y = np.divide(params, timings)
    plt.figure()

    if not args.values:
        plt.axes().yaxis.set_major_formatter(plt.NullFormatter())

    plt.ylabel("%s / timing" % args.key)
    title_final = "%s %s benchmark throughput results\n%s" % \
                  (routine, bench_variant, title)
    outpath_final = os.path.join(args.outdir, "%s_%s_%s%s" % \
                    (routine, args.plot, bench_variant, outpath))
    return y, title_final, outpath_final


def finishPlot(x, y, title, outpath, x_scale, plotted_ifuncs):
    """Finish generating current Figure.

    Args:
        x: x-axis values
        y: y-axis values
        title: figure title
        outpath: output file path
        x_scale: x-axis scale
        plotted_ifuncs: names of ifuncs to plot
    """
    plt.xlabel(args.key)
    plt.xscale(x_scale)
    plt.title(title)

    plt.grid(color="k", linestyle=args.grid, linewidth=0.5, alpha=0.5)

    for i in range(len(plotted_ifuncs)):
        plt.plot(x, y[i], marker=markers[i % len(markers)],
                 label=plotted_ifuncs[i])

    plt.legend(loc="best", fontsize="small")
    plt.savefig("%s_%s.%s" % (outpath, x_scale, args.extension),
                format=args.extension, dpi=args.resolution)

    if args.display:
        plt.show()

    plt.close()


def plotRecursive(json_iter, routine, ifuncs, bench_variant, title, outpath,
                  x_scale):
    """Plot benchmark timings.

    Args:
        json_iter: reference to json object
        routine: benchmarked string routine name
        ifuncs: names of ifuncs tested
        bench_variant: top-level benchmark variant name
        title: figure's title (generated so far)
        outpath: output file path (generated so far)
        x_scale: x-axis scale
    """

    # RECURSIVE CASE: 'variants' array found
    if "variants" in json_iter:
        # Continue recursive search for 'results' array. Record the
        # benchmark variant (configuration) in order to customize
        # the title, filename and X-axis scale for the generated figure.
        for variant in json_iter["variants"]:
            new_title = "%s%s, " % (title, variant["name"])
            new_outpath = "%s_%s" % (outpath, variant["name"].replace(" ", "_"))
            new_x_scale = "log" if variant["name"] in log_variants else x_scale

            plotRecursive(variant, routine, ifuncs, bench_variant, new_title,
                          new_outpath, new_x_scale)
        return

    # BASE CASE: 'results' array found
    domain = []
    timings = defaultdict(list)

    # Collect timings
    for result in json_iter["results"]:
        domain.append(result[args.key])
        timings[result[args.key]].append(result["timings"])

    domain = np.unique(np.array(domain))
    averages = []

    # Compute geometric mean if there are multple timings for each
    # parameter value.
    for parameter in domain:
        averages.append(gmean(timings[parameter]))

    averages = np.array(averages).transpose()

    # Choose ifuncs to plot
    if isinstance(args.ifuncs, str):
        plotted_ifuncs = ifuncs
    else:
        plotted_ifuncs = [x.replace("__", "") for x in args.ifuncs]

    plotted_indices = [ifuncs.index(x) for x in plotted_ifuncs]
    plotted_vals = averages[plotted_indices,:]

    # Plotting logic specific to each plot type
    if args.plot == "time":
        codomain, title, outpath = plotTime(plotted_vals, routine,
                                   bench_variant, title, outpath)
    elif args.plot == "rel":
        codomain, title, outpath = plotRelative(plotted_vals, averages, routine,
                                   ifuncs, bench_variant, title, outpath)
    elif args.plot == "max":
        codomain, title, outpath = plotMax(plotted_vals, routine,
                                   bench_variant, title, outpath)
    elif args.plot == "thru":
        codomain, title, outpath = plotThroughput(plotted_vals, domain, routine,
                                   bench_variant, title, outpath)

    # Plotting logic shared between plot types
    finishPlot(domain, codomain, title, outpath, x_scale, plotted_ifuncs)


def main(args):
    """Program Entry Point.

    Args:
      args: command line arguments (excluding program name)
    """

    # Select non-GUI matplotlib backend if interactive display is disabled
    if not args.display:
        mpl.use("Agg")

    global plt
    import matplotlib.pyplot as plt

    schema = None

    with open(args.schema, "r") as f:
        schema = json.load(f)

    for filename in args.bench:
        bench = None

        with open(filename, "r") as f:
            bench = json.load(f)

        validator.validate(bench, schema)

        for function in bench["functions"]:
            bench_variant = bench["functions"][function]["bench-variant"]
            ifuncs = bench["functions"][function]["ifuncs"]
            ifuncs = [x.replace("__", "") for x in ifuncs]

            plotRecursive(bench["functions"][function], function, ifuncs,
                          bench_variant, "", "", args.logarithmic)


""" main() """
if __name__ == "__main__":

    parser = argparse.ArgumentParser(description=
            "Plot string microbenchmark results",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Required parameter
    parser.add_argument("bench", nargs="+",
                        help="benchmark results file(s) in json format")

    # Optional parameters
    parser.add_argument("-b", "--baseline", type=str,
                        help="baseline ifunc for 'rel' plot")
    parser.add_argument("-d", "--display", action="store_true",
                        help="display figures")
    parser.add_argument("-e", "--extension", type=str, default="png",
                        choices=["png", "pdf", "svg"],
                        help="output file(s) extension")
    parser.add_argument("-g", "--grid", action="store_const", default="",
                        const="-", help="show grid lines")
    parser.add_argument("-i", "--ifuncs", nargs="+", default="all",
                        help="ifuncs to plot")
    parser.add_argument("-k", "--key", type=str, default="length",
                        help="key to access the varied parameter")
    parser.add_argument("-l", "--logarithmic", action="store_const",
                        default="linear", const="log",
                        help="use logarithmic x-axis scale")
    parser.add_argument("-o", "--outdir", type=str, default=os.getcwd(),
                        help="output directory")
    parser.add_argument("-p", "--plot", type=str, default="time",
                        choices=["time", "rel", "max", "thru"],
                        help="plot absolute timings, relative timings, " \
                        "performance relative to max, or throughput")
    parser.add_argument("-r", "--resolution", type=int, default=100,
                        help="dpi resolution for the generated figures")
    parser.add_argument("-s", "--schema", type=str,
                        default=os.path.join(os.path.dirname(
                        os.path.realpath(__file__)),
                        "benchout_strings.schema.json"),
                        help="schema file to validate the results file.")
    parser.add_argument("-t", "--threshold", type=int, default=5,
                        help="threshold to mark in 'rel' graph (in %%)")
    parser.add_argument("-v", "--values", action="store_true",
                        help="show actual values")

    args = parser.parse_args()
    main(args)



#!/usr/bin/python3
# Copyright (C) 2014-2021 Free Software Foundation, Inc.
# This file is part of the GNU C Library.
#
# The GNU C Library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# The GNU C Library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the GNU C Library; if not, see
# <https://www.gnu.org/licenses/>.

"""Benchmark program generator script

This script takes a function name as input and generates a program using
an input file located in the benchtests directory.  The name of the
input file should be of the form foo-inputs where 'foo' is the name of
the function.
"""

from __future__ import print_function
import sys
import os
import itertools

# Macro definitions for functions that take no arguments.  For functions
# that take arguments, the STRUCT_TEMPLATE, ARGS_TEMPLATE and
# VARIANTS_TEMPLATE are used instead.
DEFINES_TEMPLATE = '''
#define CALL_BENCH_FUNC(v, i) %(func)s ()
#define NUM_VARIANTS (1)
#define NUM_SAMPLES(v) (1)
#define VARIANT(v) FUNCNAME "()"
'''

# Structures to store arguments for the function call.  A function may
# have its inputs partitioned to represent distinct performance
# characteristics or distinct flavors of the function.  Each such
# variant is represented by the _VARIANT structure.  The ARGS structure
# represents a single set of arguments.
STRUCT_TEMPLATE = '''
#define CALL_BENCH_FUNC(v, i, x) %(func)s (x %(func_args)s)

struct args
{
%(args)s
  double timing;
};

struct _variants
{
  const char *name;
  int count;
  struct args *in;
};
'''

# The actual input arguments.
ARGS_TEMPLATE = '''
struct args in%(argnum)d[%(num_args)d] = {
%(args)s
};
'''

# The actual variants, along with macros defined to access the variants.
VARIANTS_TEMPLATE = '''
struct _variants variants[%(num_variants)d] = {
%(variants)s
};

#define NUM_VARIANTS %(num_variants)d
#define NUM_SAMPLES(i) (variants[i].count)
#define VARIANT(i) (variants[i].name)
'''

# Epilogue for the generated source file.
EPILOGUE = '''
#define RESULT(__v, __i) (variants[(__v)].in[(__i)].timing)
#define RESULT_ACCUM(r, v, i, old, new) \\
        ((RESULT ((v), (i))) = (RESULT ((v), (i)) * (old) + (r)) / ((new) + 1))
#define BENCH_FUNC(i, j) ({%(getret)s CALL_BENCH_FUNC (i, j, );})
#define BENCH_FUNC_LAT(i, j) ({%(getret)s CALL_BENCH_FUNC (i, j, %(latarg)s);})
#define BENCH_VARS %(defvar)s
#define FUNCNAME "%(func)s"
#include "bench-skeleton.c"'''


def gen_source(func, directives, all_vals):
    """Generate source for the function

    Generate the C source for the function from the values and
    directives.

    Args:
      func: The function name
      directives: A dictionary of directives applicable to this function
      all_vals: A dictionary input values
    """
    # The includes go in first.
    for header in directives['includes']:
        print('#include <%s>' % header)

    for header in directives['include-sources']:
        print('#include "%s"' % header)

    # Print macros.  This branches out to a separate routine if
    # the function takes arguments.
    if not directives['args']:
        print(DEFINES_TEMPLATE % {'func': func})
        outargs = []
    else:
        outargs = _print_arg_data(func, directives, all_vals)

    # Print the output variable definitions if necessary.
    for out in outargs:
        print(out)

    # If we have a return value from the function, make sure it is
    # assigned to prevent the compiler from optimizing out the
    # call.
    getret = ''
    latarg = ''
    defvar = ''

    if directives['ret']:
        print('static %s volatile ret;' % directives['ret'])
        print('static %s zero __attribute__((used)) = 0;' % directives['ret'])
        getret = 'ret = func_res = '
        # Note this may not work if argument and result type are incompatible.
        latarg = 'func_res * zero +'
        defvar = '%s func_res = 0;' % directives['ret']

    # Test initialization.
    if directives['init']:
        print('#define BENCH_INIT %s' % directives['init'])

    print(EPILOGUE % {'getret': getret, 'func': func, 'latarg': latarg, 'defvar': defvar })


def _print_arg_data(func, directives, all_vals):
    """Print argument data

    This is a helper function for gen_source that prints structure and
    values for arguments and their variants and returns output arguments
    if any are found.

    Args:
      func: Function name
      directives: A dictionary of directives applicable to this function
      all_vals: A dictionary input values

    Returns:
      Returns a list of definitions for function arguments that act as
      output parameters.
    """
    # First, all of the definitions.  We process writing of
    # CALL_BENCH_FUNC, struct args and also the output arguments
    # together in a single traversal of the arguments list.
    func_args = []
    arg_struct = []
    outargs = []

    for arg, i in zip(directives['args'], itertools.count()):
        if arg[0] == '<' and arg[-1] == '>':
            pos = arg.rfind('*')
            if pos == -1:
                die('Output argument must be a pointer type')

            outargs.append('static %s out%d __attribute__((used));' % (arg[1:pos], i))
            func_args.append(' &out%d' % i)
        else:
            arg_struct.append('  %s volatile arg%d;' % (arg, i))
            func_args.append('variants[v].in[i].arg%d' % i)

    print(STRUCT_TEMPLATE % {'args' : '\n'.join(arg_struct), 'func': func,
                             'func_args': ', '.join(func_args)})

    # Now print the values.
    variants = []
    for (k, vals), i in zip(all_vals.items(), itertools.count()):
        out = ['  {%s, 0},' % v for v in vals]

        # Members for the variants structure list that we will
        # print later.
        variants.append('  {"%s", %d, in%d},' % (k, len(vals), i))
        print(ARGS_TEMPLATE % {'argnum': i, 'num_args': len(vals),
                               'args': '\n'.join(out)})

    # Print the variants and the last set of macros.
    print(VARIANTS_TEMPLATE % {'num_variants': len(all_vals),
                               'variants': '\n'.join(variants)})
    return outargs


def _process_directive(d_name, d_val):
    """Process a directive.

    Evaluate the directive name and value passed and return the
    processed value. This is a helper function for parse_file.

    Args:
      d_name: Name of the directive
      d_val: The string value to process

    Returns:
      The processed value, which may be the string as it is or an object
      that describes the directive.
    """
    # Process the directive values if necessary.  name and ret don't
    # need any processing.
    if d_name.startswith('include'):
        d_val = d_val.split(',')
    elif d_name == 'args':
        d_val = d_val.split(':')

    # Return the values.
    return d_val


def parse_file(func):
    """Parse an input file

    Given a function name, open and parse an input file for the function
    and get the necessary parameters for the generated code and the list
    of inputs.

    Args:
      func: The function name

    Returns:
      A tuple of two elements, one a dictionary of directives and the
      other a dictionary of all input values.
    """
    all_vals = {}
    # Valid directives.
    directives = {
            'name': '',
            'args': [],
            'includes': [],
            'include-sources': [],
            'ret': '',
            'init': ''
    }

    try:
        with open('%s-inputs' % func) as f:
            for line in f:
                # Look for directives and parse it if found.
                if line.startswith('##'):
                    try:
                        d_name, d_val = line[2:].split(':', 1)
                        d_name = d_name.strip()
                        d_val = d_val.strip()
                        directives[d_name] = _process_directive(d_name, d_val)
                    except (IndexError, KeyError):
                        die('Invalid directive: %s' % line[2:])

                # Skip blank lines and comments.
                line = line.split('#', 1)[0].rstrip()
                if not line:
                    continue

                # Otherwise, we're an input.  Add to the appropriate
                # input set.
                cur_name = directives['name']
                all_vals.setdefault(cur_name, [])
                all_vals[cur_name].append(line)
    except IOError as ex:
        die("Failed to open input file (%s): %s" % (ex.filename, ex.strerror))

    return directives, all_vals


def die(msg):
    """Exit with an error

    Prints an error message to the standard error stream and exits with
    a non-zero status.

    Args:
      msg: The error message to print to standard error
    """
    print('%s\n' % msg, file=sys.stderr)
    sys.exit(os.EX_DATAERR)


def main(args):
    """Main function

    Use the first command line argument as function name and parse its
    input file to generate C source that calls the function repeatedly
    for the input.

    Args:
      args: The command line arguments with the program name dropped

    Returns:
      os.EX_USAGE on error and os.EX_OK on success.
    """
    if len(args) != 1:
        print('Usage: %s <function>' % sys.argv[0])
        return os.EX_USAGE

    directives, all_vals = parse_file(args[0])
    gen_source(args[0], directives, all_vals)
    return os.EX_OK


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
